"""
INITIAL_SCREEN - Rapid screening of univariate variance component models
INITIAL_SCREEN screens known covariates for possible filtering before
applying the lvreml function. Screening is based on estimating the
variance explained by each covariate alone, followed by a pruning step to 
obtain a set of non-redundant, linearly independent set of covariates. 
Details and rationale are in Section S4 of the paper.

USAGE: [beta2,varexpl] = initial_screen(C,Z)
       [beta2,varexpl,idx] = initial_screen(C,Z,theta)
INPUT: C - (n x n) empirical sample covariance matrix of expression data, 
           see also data_prep
       Z - (n x d) matrix of normalized data for d covariates (known
           confounders) in n samples, see also data_prep
OUTPUT: beta2   - (d x 1) vector, variance parameter in univariate lvreml
                  model for each covariate, see Section S4 of the paper
        varexpl - (d x 1) vector, variance in C explained by each
                  covariate (= beta2/tr(C))
        idx     - indices for a subset of linearly independent covariates
                  with varexpl>=theta, only returned if number of input
                  arguments equals 3
                  
AUTHOR: Muhammad Ammar Malik
        muhammad.malik@uib.no
        https://ammarmalik93.github.io/
REFERENCE: MA Malik and T Michoel. Restricted maximum-likelihood method
for learning latent variance components in gene expression data with
known and unknown confounders. 
"""
import numpy as np
from numpy.linalg import matrix_rank

def initial_screen(C,Z,vcut,verbose):
    ns,nc = Z.shape # no. of covariates and samples
    beta2 = np.zeros((nc,),dtype=float)
    varexpl = np.zeros((nc,),dtype=float)
    for k in range(nc):
        beta2[k] = (ns*(np.dot(np.dot(Z[:,k].T,C),Z[:,k]))/(ns-1)) - (np.trace(C)/(ns-1))
    beta2[beta2<0] = 0

    varexpl[beta2>0] = beta2[beta2>0]/np.trace(C)

#    varexpl = varexpl[varexpl>0]
    # get best linearly independent set with varexpl>threshold, if necessary
    vs = np.sort(varexpl)[::-1]
    t = np.argsort(varexpl)[::-1]
    if vcut == 0:
        if verbose == True:
            print("No covariates to be selected") 
        idx = np.array([], dtype=int)
    elif vcut < 1 and vcut > 0:
 #       print("Finding linearly independent covariates w.r.t given threshold %.4f"%vcut)   
        idx = t[vs>vcut]
        for k in range(1,len(idx)):
            ID = idx[0:k+1]
            rnk = matrix_rank(Z[:,ID[ID>0]])
            if rnk < (ID>0).sum():
                idx[k] = 0
        idx = idx[idx>0]
        
    else:
        if verbose == True:
            print("Finding best %d linearly independent covariates"%vcut)
        idx = t
        temp = 1
        for k in range(1,len(idx)):
            ID = idx[0:k+1]
            rnk = matrix_rank(Z[:,ID[ID>0]])
            if rnk < (ID>0).sum():
                idx[k] = 0
            else:
                temp = temp+1
                if temp == vcut:
                    break
        idx = idx[idx>0]
        idx = idx[0:vcut]

    
    return beta2,varexpl,idx



"""
DATA_PREP - Prepare expression and known covariate data for use in lvreml functions
DATA_PREP prepares the expression and known covariate data for use in
other functions of the lvreml package. Details and rationale are in
Section S2 of the paper.
 
USAGE: [C,Yn,Zn] = data_prep(Y,Z)

INPUT: Y - (n x m) matrix of expression data for m genes in n samples
       Z - (n x d) matrix of data for d covariates (known confounders) in
           n samples

OUTPUT: C  - empirical sample covariance matrix for the expression data,
        Yn - centred expression data (all samples centred to have mean zero)   
        Zn - normalized covariate data (all variables scaled to have unit
             L2 norm)
        
AUTHOR: Muhammad Ammar Malik
        muhammad.malik@uib.no
        https://ammarmalik93.github.io/

REFERENCE: MA Malik and T Michoel. Restricted maximum-likelihood method
for learning latent variance components in gene expression data with
known and unknown confounders.

"""
import numpy as np
import warnings


def data_prep(Y,Z):
# Check that data is in right format (columns are variables, rows are samples)    
    if Z.size != 0:
        if Y.shape[0] != Z.shape[0]:
            # rows are not matching, check columns
            if Y.shape[1] == Z.shape[1]:
                # columns are samples -> transpose both
                Y = Y.T
                Z = Z.T
            else:
                raise Exception('Cannot determine sample dimension')
    else:
        warnings.warn("No covariates provided, assuming expression data is in format samples x genes")
    
    # Center Y to remove fixed effects on mean and get overlap matrix         
    Yn = Y - np.array(Y.mean(axis=1)).reshape(Y.shape[0],1)
    C = np.dot(Yn,Yn.T)/Y.shape[1]
    
    if Z.size != 0:

        Zn = (Z.T/np.sqrt(np.sum(np.square(Z),axis=0)).reshape(Z.shape[1],1)).T
    else:
        Zn = np.empty((0,0))
    return C,Zn,Yn



from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="LVREML",
    version="0.1.0",
    author="Muhammad Ammar Malik",
    author_email="muhammad.malik@uib.no",
    description="LVREML - Restricted maximum-likelihood solution for linear mixed models with known and latent variance components LVREML computes the restricted maximum-likelihood solution for linear mixed models with known and latent variance components",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/AmmarMalik93/LVREML-Python",
 #   project_urls={
 #       "Bug Tracker": "",
 #   },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=['lvreml']
)



"""
LVREML - Restricted maximum-likelihood solution for linear mixed models with known and latent variance components
LVREML computes the restricted maximum-likelihood solution for linear mixed 
models with known and latent variance components  Details and rationale are in
Sections S2, S4-S7 of the paper.

IMPORTANT NOTE: Input parameters C and Z must be scaled and normalized
using the "data_prep" function to ensure correct output!

USAGE: [X,alpha2,B,D,sigma2,K] = lvreml(C,Z,targetX)

INPUT: C       - (n x n) sample covariance matrix of expression data for
                 m genes in n samples, see also data_prep
       Z       - (n x d) matrix of L1-normalized data for d covariates
                 (known confounders) in n samples, see also data_prep
       targetX - number, interpreted as target number of latent
                 variables (targetX >= 1), or target value for the total
                 variance in Y explained by the model (0<=targetX<1)
OUTPUT: X      - (n x p) matrix of latent variable data (rows are
                 samples, columns are variables, p is automatically 
                 determined), also called X in the paper.
        alpha2 - (p x 1) vector of variance explained by each latent
                 variable; this is the diagonal of the diagonal matrix A
                 in the paper.  
        B      - (d x d) matrix of covariances among the known
                 covariates, also called B in the paper.
        D      - (d x p) matrix of covariances between known and latent
                 variables, also called D in the paper.
        sigma2 - the squared residual variance
        K      - the restricted maximum-likelihood sample covariance
                 matrix
                 
AUTHOR: Muhammad Ammar Malik
        muhammad.malik@uib.no
        https://ammarmalik93.github.io/

REFERENCE: MA Malik and T Michoel. Restricted maximum-likelihood method
for learning latent variance components in gene expression data with
known and unknown confounders. 
"""

import numpy as np


def lvreml(C,Z,targetX):
    # LVREML - Restricted maximum-likelihood solution
    # for linear mixed models with known and latent
    # variance components
    
    # C,Z,_ = data_prep(Y,Z)
    ns = len(C) # number of samples
    trC = np.trace(C)
    
    # Full SVD of known covariates
    if Z.size != 0:
        U,G,V = np.linalg.svd(Z)
        V = V.T
        nc = G.shape[0] # number of linearly independent covariates
        
        # Split in space spanned by covariates and orthogonal complement
        G1 = G*np.eye(nc)
        V1 = V[:,0:nc]
        U1 = U[:,0:nc]
        U2 = U[:, nc:]

        
        # Corresponding matrix blocks
        C11 = (U1.T).dot(C).dot(U1)
        C22 = (U2.T).dot(C).dot(U2)
        EC11,_ = np.linalg.eigh(C11)
        
        # Get REML estimate for X       
        if C22.size != 0:
            Ex,Vx = np.linalg.eigh(C22)
            Evx = np.sort(Ex)
            t = np.argsort(Ex)
            # minimum eigenvalue on known covariate space, needed to
            # check validitiy of the solution
            lambdamin = EC11.min()
            # variance not explained by X, at all possible number of 
            # latent variables; this is called "f(p)" in the appendix
            # of the paper
            sigma2vec = np.cumsum(Evx)/range(1,Evx.shape[0]+1)
#            sigma2vec = np.round(sigma2vec,4)
            # check if 3rd argument wants a given number of latent
            # variables or target variance explained; in both cases
            # adjust to make sure the solution will be valid
            if type(targetX) == np.int32: # >= 1:
                nx = np.round(targetX) # number of latent variables
                # find the smallest number of latent variables we must
                # include (largest number we can cut)
                ncut_max = (np.array(np.logical_and(sigma2vec<lambdamin,Evx>Evx[0]).nonzero()).T[-1][0])
                # cut fewer if we've asked for more latent variables
                if ncut_max < Evx.shape[0]-nx:
                    print("Number of latent variables asked for is too small")
                    print("Infering minimum number of latent variables: %d"%(Evx.shape[0]-ncut_max-1))
                    ncut = ncut_max#+1
                elif nx > Evx.shape[0]:
                    print("Number of latent variables asked for is too large")
                    print("Infering maximum number of latent variables: %d"%Evx.shape[0])
                    ncut = np.array([])
                else:
                    print("Inferring desired number of latent variables: %d"%(nx))
                    ncut = Evx.shape[0]-nx
 
            elif type(targetX) == np.float64: #targetX >= 0:
                varexpl = targetX # target variance explained
                # Set the target residual variance
                resvar = min([(1-varexpl)*trC/ns, lambdamin])
                #resvar = np.round(resvar,4)
                # find number of variables that need to be cut to explain
                # no less than the target variance explained
                ncut = (np.array(np.logical_and(sigma2vec<resvar,Evx>Evx[0]).nonzero()))#.T[-1][0]+1
                if ncut.size == 0:
                    
                    ncut = np.array([])
                else:
                    
                    ncut = ncut.T[-1][0]+1
                
            else:
                raise Exception('lvreml::lvreml::3rd argument must be integer or value between zero and one')
            
            if ncut.size == 0:
                ncut = 0
                sigma2 = 0
            else:
                # estimate sigma^2 parameter
                sigma2 = Evx[0:ncut].mean() #sigma2 = Evx[0:ncut+1].mean()
            
            # pull back selected eigenvectors of C22 to original space
            X = np.dot(U2,Vx[:,np.flip(t[ncut:])]);
            alpha2 = np.flip(Evx[ncut:]-sigma2)
            
        else:
            X = np.array([])
            alpha2 = np.array([])
            sigma2 = 0
        
        # Get estimates for the covaraicne matrix among the known
        # confounders
        V1G = (V1).dot(np.linalg.inv(G1))
        B = (V1G).dot(C11-sigma2*np.eye(nc)).dot(V1G.T)
        
        # Get estimates for the covariance matrix between the known
        # confounders and latent variables
        
        if X.size != 0:
            D = (V1G).dot(U1.T).dot(C).dot(X)
        else:
            D = np.array([])
        
        # Return covariance matrix K
        if X.size != 0:
            K = np.c_[Z,X].dot((np.r_[(np.c_[B,D]),(np.c_[D.T, np.diag(alpha2)])])).dot(np.r_[Z.T,X.T]) + np.eye(ns)*sigma2
        else:
            K = Z.dot(B.dot(Z.T)) + np.eye(ns)*sigma2
        K = 0.5*(K+K.T)
        
    else:
        # no covariates provided, only estimating hidden components
        # using PCA
        print("no covariates provided, only estimating hidden components using PCA ")
        B = np.array([])
        D = np.array([])
        Ex,Vx = np.linalg.eigh(C)
        Evx = np.sort(Ex)
        t = np.argsort(Ex)
        Vx = Vx[:,t]
        sigma2vec = np.cumsum(Evx)/range(1,Evx.shape[0]+1)
           
        nmax = np.array((Evx[1:]-sigma2vec[:len(sigma2vec)-1]>0).nonzero()).T[-1][0]
        if type(targetX) == np.int32 : # >= 1:
            nx = ns - np.round(targetX)
        elif type(targetX) == np.float64 : # >= 0:
            varexpl = targetX        
            resvar = (1-varexpl)*trC/ns
            nx = np.array((sigma2vec<resvar).nonzero()).T[-1][0]+1
        else:
            raise Exception('lvreml::lvreml::3rd argument must be integer or value between zero and one')
        
            

        nx = min([nx,nmax])
        sigma2 = Evx[0:nx].mean()
        alpha2 = np.flip(Evx[-(len(Evx)-nx):]-sigma2)
        X = np.fliplr(Vx[:,-(len(Evx)-nx):])
        
        # Return covariance matrix K
        K = np.dot(np.dot(X,np.diag(alpha2)),X.T) +np.eye(ns)*sigma2
        K = 0.5*(K+K.T)
        
    return X,alpha2,B,D,sigma2,K
